"""
Offset: basic force verification.

This simple verification check ensures that basic forces are correctly
assigned using a single beamcolumn element in a variety of
arrangements.

Written: Thu Nov 28 06:45:08 PM PST 2024

"""

from __future__ import annotations

import numpy as np

from osmg.analysis.common import PointLoad
from osmg.analysis.load_case import LoadCaseRegistry
from osmg.analysis.supports import FixedSupport
from osmg.core.model import Model2D, Model3D
from osmg.creators.component import BeamColumnCreator
from osmg.graphics.plotly import (
    BasicForceConfiguration,
    DeformationConfiguration,
    Figure3D,
    Figure3DConfiguration,
)
from osmg.model_objects.node import Node
from osmg.model_objects.section import ElasticSection


def no_offset_2d() -> None:
    """Axially loaded cantilever column, 2D, no offset."""
    column_model = Model2D(name='Test model', dimensionality='2D Frame')
    n1 = Node(uid_generator=column_model.uid_generator, coordinates=((0.00, 0.00)))
    n2 = Node(uid_generator=column_model.uid_generator, coordinates=((0.00, 5.00)))
    column_model.nodes.add(n1)
    column_model.nodes.add(n2)
    simple_section = ElasticSection(
        column_model.uid_generator,
        'Test Section',
        e_mod=1e3,
        area=1e3,
        i_y=1.00,
        i_x=1.00,
        g_mod=1.00,
        j_mod=1.00,
        sec_w=0.00,
    )
    bcg = BeamColumnCreator(column_model, 'elastic')
    bcg.generate_plain_component_assembly(
        tags={'column'},
        node_i=n2,
        node_j=n1,
        n_sub=1,
        eo_i=np.array((0.00, 0.00)),
        eo_j=np.array((0.00, 0.00)),
        section=simple_section,
        transf_type='Linear',
    )
    load_case_registry = LoadCaseRegistry(column_model)
    fixed_support = FixedSupport((True, True, True))
    load_case_registry.dead['test'].fixed_supports[n1.uid] = fixed_support
    load_case_registry.dead['test'].load_registry.nodal_loads[n2.uid] = PointLoad(
        (0.00, -10.00, 0.00)  # lb
    )
    load_case_registry.run()

    axial_df, shear_y_df, shear_z_df, torsion_df, moment_y_df, moment_z_df = (
        load_case_registry.dead['test'].calculate_basic_forces(
            'default_basic_force',
            column_model.components.get_line_element_lengths(),
            ndm=2,
            num_stations=12,
        )
    )
    assert np.allclose(axial_df.to_numpy(), -10.00)
    assert np.allclose(shear_y_df.to_numpy(), 0.00)
    assert np.allclose(shear_z_df.to_numpy(), 0.00)
    assert np.allclose(torsion_df.to_numpy(), 0.00)
    assert np.allclose(moment_y_df.to_numpy(), 0.00)
    assert np.allclose(moment_z_df.to_numpy(), 0.00)

    displacements = (
        load_case_registry.dead['test'].analysis.recorders['default_node'].get_data()
    )
    assert np.allclose(
        displacements.to_numpy(), np.array((0.0, 0.0, 0.0, 0.0, -5.0e-05, 0.0))
    )
    # plot(column_model, load_case_registry)  # appears correct.


def offset_2d() -> None:
    """Axially loaded cantilever column, 2D, offset."""
    column_model = Model2D(name='Test model', dimensionality='2D Frame')
    n1 = Node(uid_generator=column_model.uid_generator, coordinates=((0.00, 0.00)))
    n2 = Node(uid_generator=column_model.uid_generator, coordinates=((0.00, 5.00)))
    column_model.nodes.add(n1)
    column_model.nodes.add(n2)
    simple_section = ElasticSection(
        column_model.uid_generator,
        'Test Section',
        e_mod=1e3,
        area=1e3,
        i_y=1.00,
        i_x=1.00,
        g_mod=1.00,
        j_mod=1.00,
        sec_w=0.00,
    )
    bcg = BeamColumnCreator(column_model, 'elastic')
    bcg.generate_plain_component_assembly(
        tags={'column'},
        node_i=n2,
        node_j=n1,
        n_sub=1,
        eo_i=np.array((1.00, 0.00)),
        eo_j=np.array((1.00, 0.00)),
        section=simple_section,
        transf_type='Linear',
    )
    load_case_registry = LoadCaseRegistry(column_model)
    fixed_support = FixedSupport((True, True, True))
    load_case_registry.dead['test'].fixed_supports[n1.uid] = fixed_support
    load_case_registry.dead['test'].load_registry.nodal_loads[n2.uid] = PointLoad(
        (0.00, -10.00, 0.00)  # lb
    )
    load_case_registry.run()

    axial_df, shear_y_df, shear_z_df, torsion_df, moment_y_df, moment_z_df = (
        load_case_registry.dead['test'].calculate_basic_forces(
            'default_basic_force',
            column_model.components.get_line_element_lengths(),
            ndm=2,
            num_stations=12,
        )
    )
    assert np.allclose(axial_df.to_numpy(), -10.00)
    assert np.allclose(shear_y_df.to_numpy(), 0.00)
    assert np.allclose(shear_z_df.to_numpy(), 0.00)
    assert np.allclose(torsion_df.to_numpy(), 0.00)
    assert np.allclose(moment_y_df.to_numpy(), 0.00)
    assert np.allclose(moment_z_df.to_numpy(), -10.00)

    displacements = (
        load_case_registry.dead['test'].analysis.recorders['default_node'].get_data()
    )
    assert np.allclose(
        displacements.to_numpy(), np.array((0.0, 0.0, 0.0, -0.125, -0.05005, 0.05))
    )
    # plot(column_model, load_case_registry)  # appears correct.


def no_offset_3d() -> None:
    """Axially loaded cantilever column, 3D, no offset."""
    column_model = Model3D(name='Test model', dimensionality='3D Frame')
    n1 = Node(
        uid_generator=column_model.uid_generator, coordinates=((0.00, 0.00, 0.00))
    )
    n2 = Node(
        uid_generator=column_model.uid_generator, coordinates=((0.00, 0.00, 5.00))
    )
    column_model.nodes.add(n1)
    column_model.nodes.add(n2)
    simple_section = ElasticSection(
        column_model.uid_generator,
        'Test Section',
        e_mod=1e3,
        area=1e3,
        i_y=1.00,
        i_x=1.00,
        g_mod=1.00,
        j_mod=1.00,
        sec_w=0.00,
    )
    bcg = BeamColumnCreator(column_model, 'elastic')
    bcg.generate_plain_component_assembly(
        tags={'column'},
        node_i=n2,
        node_j=n1,
        n_sub=1,
        eo_i=np.array((0.00, 0.00, 0.0)),
        eo_j=np.array((0.00, 0.00, 0.0)),
        section=simple_section,
        transf_type='Linear',
        angle=90.00 / 360.00 * 2.00 * np.pi,
    )
    load_case_registry = LoadCaseRegistry(column_model)
    fixed_support = FixedSupport((True, True, True, True, True, True))
    load_case_registry.dead['test'].fixed_supports[n1.uid] = fixed_support
    load_case_registry.dead['test'].load_registry.nodal_loads[n2.uid] = PointLoad(
        (0.00, 0.00, -10.00, 0.00, 0.00, 0.00)  # lb
    )
    load_case_registry.run()
    axial_df, shear_y_df, shear_z_df, torsion_df, moment_y_df, moment_z_df = (
        load_case_registry.dead['test'].calculate_basic_forces(
            'default_basic_force',
            column_model.components.get_line_element_lengths(),
            ndm=3,
            num_stations=12,
        )
    )
    assert np.allclose(axial_df.to_numpy(), -10.00)
    assert np.allclose(shear_y_df.to_numpy(), 0.00)
    assert np.allclose(shear_z_df.to_numpy(), 0.00)
    assert np.allclose(torsion_df.to_numpy(), 0.00)
    assert np.allclose(moment_y_df.to_numpy(), 0.00)
    assert np.allclose(moment_z_df.to_numpy(), 0.00)

    displacements = (
        load_case_registry.dead['test'].analysis.recorders['default_node'].get_data()
    )
    assert np.allclose(
        displacements.to_numpy(),
        np.array((0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -5.0e-05, 0.0, 0.0, 0.0)),
    )
    plot(column_model, load_case_registry)  # appears correct.


def offset_3d() -> None:
    """Axially loaded cantilever column, 3D, offset."""
    column_model = Model3D(name='Test model', dimensionality='3D Frame')
    n1 = Node(
        uid_generator=column_model.uid_generator, coordinates=((0.00, 0.00, 0.00))
    )
    n2 = Node(
        uid_generator=column_model.uid_generator, coordinates=((0.00, 0.00, 5.00))
    )
    column_model.nodes.add(n1)
    column_model.nodes.add(n2)
    simple_section = ElasticSection(
        column_model.uid_generator,
        'Test Section',
        e_mod=1e3,
        area=1e3,
        i_y=1.00,
        i_x=1.00,
        g_mod=1.00,
        j_mod=1.00,
        sec_w=0.00,
    )
    bcg = BeamColumnCreator(column_model, 'elastic')
    bcg.generate_plain_component_assembly(
        tags={'column'},
        node_i=n2,
        node_j=n1,
        n_sub=1,
        eo_i=np.array((1.00, 0.00, 0.0)),
        eo_j=np.array((1.00, 0.00, 0.0)),
        section=simple_section,
        transf_type='Linear',
        angle=90.00 / 360.00 * 2.00 * np.pi,
    )
    load_case_registry = LoadCaseRegistry(column_model)
    fixed_support = FixedSupport((True, True, True, True, True, True))
    load_case_registry.dead['test'].fixed_supports[n1.uid] = fixed_support
    load_case_registry.dead['test'].load_registry.nodal_loads[n2.uid] = PointLoad(
        (0.00, 0.00, -10.00, 0.00, 0.00, 0.00)  # lb
    )
    load_case_registry.run()
    axial_df, shear_y_df, shear_z_df, torsion_df, moment_y_df, moment_z_df = (
        load_case_registry.dead['test'].calculate_basic_forces(
            'default_basic_force',
            column_model.components.get_line_element_lengths(),
            ndm=3,
            num_stations=12,
        )
    )
    assert np.allclose(axial_df.to_numpy(), -10.00)
    assert np.allclose(shear_y_df.to_numpy(), 0.00)
    assert np.allclose(shear_z_df.to_numpy(), 0.00)
    assert np.allclose(torsion_df.to_numpy(), 0.00)
    assert np.allclose(moment_y_df.to_numpy(), 0.00)
    assert np.allclose(moment_z_df.to_numpy(), 10.00)

    displacements = (
        load_case_registry.dead['test'].analysis.recorders['default_node'].get_data()
    )
    assert np.allclose(
        displacements.to_numpy(),
        np.array(
            (0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -0.125, 0.0, -0.05005, 0.0, -0.05, 0)
        ),
    )
    # plot(column_model, load_case_registry)  # appears correct.


def plot(model: Model2D | Model3D, load_case_registry: LoadCaseRegistry) -> None:
    """Plot the deformed shape and basic forces."""
    if isinstance(model, Model2D):
        ndm = 2
        ndf = 3
    else:
        ndm = 3
        ndf = 6
    deformation_configuration = DeformationConfiguration(
        reference_length=model.reference_length(),
        ndf=ndf,
        ndm=ndm,
        data=load_case_registry.dead['test']
        .analysis.recorders['default_node']
        .get_data(),
        step=0,
        amplification_factor=None,  # Figure it out.
    )
    basic_force_configuration = BasicForceConfiguration(
        reference_length=model.reference_length(),
        ndf=ndf,
        ndm=ndm,
        data=load_case_registry.dead['test'].calculate_basic_forces(
            'default_basic_force',
            model.components.get_line_element_lengths(),
            ndm=ndm,
            num_stations=12,
        ),
        step=-1,
        force_to_length_factor=0.1,
        moment_to_length_factor=0.1,
    )

    fig = Figure3D(Figure3DConfiguration(ndm=ndm))  # type: ignore
    fig.add_nodes(list(model.nodes.values()), 'primary', overlay=True)
    fig.add_components(list(model.components.values()), overlay=True)
    fig.add_nodes(list(model.nodes.values()), 'primary', deformation_configuration)
    fig.add_components(list(model.components.values()), deformation_configuration)
    fig.add_supports(
        model.nodes,
        load_case_registry.dead['test'].fixed_supports,
        symbol_size=12.00 / 120.0,
    )
    fig.add_udl(
        load_case_registry.dead['test'].load_registry.component_udl,
        model.components,
        force_to_length_factor=0.10,
        offset=0.00,
    )
    fig.add_loads(
        load_case_registry.dead['test'].load_registry.nodal_loads,
        model.nodes,
        force_to_length_factor=0.10,
        offset=0.0,
        head_length=24.0 / 120.0,
        head_width=24.0 / 120.0,
        base_width=5.0 / 120.0,
    )
    fig.add_basic_forces(
        components=list(model.components.values()),
        basic_force_configuration=basic_force_configuration,
    )
    fig.show()



"""Recorder objects."""

from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path
from typing import Literal

import pandas as pd

from osmg.core.uid_object import UIDObject


@dataclass
class Recorder(UIDObject):
    """Recorder base class."""

    file_name: str
    # TODO(JVM): figure out binary format.

    def __post_init__(self) -> None:
        """Post-initialization."""
        self._data = None

    def ops_args(self) -> list[object]:  # noqa: PLR6301
        """Obtain the OpenSees arguments."""
        msg = 'Child classes should implement this.'
        raise NotImplementedError(msg)

    def get_data(self) -> pd.DataFrame:  # noqa: PLR6301
        """Retrieve the data."""
        msg = 'Child classes should implement this.'
        raise NotImplementedError(msg)

    def set_data(self, data: pd.DataFrame) -> None:
        """Overwrite the data."""
        self._data = data


@dataclass
class NodeRecorder(Recorder):
    """
    OpenSees Node recorder.

    The Node recorder type records the response of a number of nodes
    at every converged step.

    Note: I haven't been able to get the `eigen` type working.

    https://opensees.berkeley.edu/wiki/index.php?title=Node_Recorder
    """

    recorder_type: Literal['Node', 'EnvelopeNode']
    nodes: tuple[int, ...]
    dofs: tuple[int, ...]
    response_type: Literal[
        'disp', 'vel', 'accel', 'incrDisp', 'reaction', 'rayleighForces'
    ]
    number_of_significant_digits: int
    output_time: bool | None
    delta_t: float | None = field(default=None)
    time_series_tag: int | None = field(default=None)

    def ops_args(self) -> list[object]:
        """
        Obtain the OpenSees arguments.

        Returns:
          The OpenSees arguments.
        """
        output: list[object] = [
            self.recorder_type,
            '-file',
            self.file_name,
            '-precision',
            self.number_of_significant_digits,
        ]
        if self.time_series_tag:
            output.extend(
                [
                    '-timeSeries',
                    self.time_series_tag,
                ]
            )
        if self.output_time:
            output.extend(['-time'])
        if self.delta_t:
            output.extend(['-dT', self.delta_t])
        output.extend(['-node', *self.nodes])
        output.extend(['-dof', *self.dofs])
        output.append(self.response_type)
        return output

    def get_data(self) -> pd.DataFrame:
        """
        Retrieve the data.

        Returns:
          The data.
        """
        if self.recorder_type == 'EnvelopeNode':
            raise NotImplementedError

        if self._data is None:
            index_col = 0 if self.output_time else None
            data = pd.read_csv(
                self.file_name,
                sep=' ',
                index_col=index_col,
                header=None,
                engine='pyarrow',
            )
            data = data.astype(float)
            header_data = [(node, dof) for node in self.nodes for dof in self.dofs]
            data.columns = pd.MultiIndex.from_tuples(
                header_data, names=('node', 'dof')
            )
            data.index.name = 'time' if self.output_time else None
            self._data = data
        return self._data


@dataclass
class DriftRecorder(Recorder):
    """
    OpenSees Drift recorder.

    The Drift type records the displacement drift between two
    nodes. The drift is taken as the ratio between the prescribed
    relative displacement and the specified distance between the
    nodes.
    https://opensees.berkeley.edu/wiki/index.php?title=Node_Recorder
    """

    nodes_i: tuple[int, ...]
    nodes_j: tuple[int, ...]
    perpendicular_directions: tuple[int, ...]
    dofs: tuple[int, ...]
    file_name: str
    number_of_significant_digits: int
    output_time: bool
    # delta_t: float | None = field(default=None)

    def ops_args(self) -> list[object]:
        """
        Obtain the OpenSees arguments.

        Returns:
          The OpenSees arguments.
        """
        output: list[object] = [
            'Drift',
            '-file',
            self.file_name,
            '-precision',
            self.number_of_significant_digits,
        ]
        if self.output_time:
            output.extend(['-time'])
        # if self.delta_t:
        #     output.extend(['-dT', self.delta_t])
        output.extend(['-iNode', *self.nodes_i])
        output.extend(['-jNode', *self.nodes_j])
        output.extend(['-dof', *self.dofs])
        output.extend(['-perpDirn', *self.perpendicular_directions])
        # TODO(JVM): see if delta_t is supported.
        return output


@dataclass
class ElementRecorder(Recorder):
    """
    OpenSees Element recorder.

    The Element recorder type records the response of a number of
    elements at every converged step. The response recorded is
    element-dependent and also depends on the arguments which are
    passed to the setResponse() element method.
    https://opensees.berkeley.edu/wiki/index.php?title=Node_Recorder
    """

    recorder_type: Literal['Element', 'EnvelopeElement']
    elements: tuple[int, ...]
    element_arguments: tuple[str, ...]
    file_name: str
    number_of_significant_digits: int | None
    output_time: bool
    delta_t: float | None = field(default=None)

    def ops_args(self) -> list[object]:
        """
        Obtain the OpenSees arguments.

        Returns:
          The OpenSees arguments.
        """
        output: list[object] = [
            self.recorder_type,
            '-file',
            self.file_name,
        ]
        if self.number_of_significant_digits:
            output.extend(
                [
                    '-precision',
                    self.number_of_significant_digits,
                ]
            )
        if self.output_time:
            output.extend(['-time'])
        if self.delta_t:
            output.extend(['-dT', self.delta_t])
        output.extend(['-ele', *self.elements])
        output.extend([*self.element_arguments])
        return output

    def get_data(self, *, update_index: bool = True) -> pd.DataFrame:
        """
        Retrieve the data.

        Returns:
          The data.
        """
        if self.recorder_type == 'EnvelopeElement':
            raise NotImplementedError

        if self._data is None:
            index_col = 0 if self.output_time else None
            data = pd.read_csv(
                self.file_name,
                sep=' ',
                index_col=index_col,
                header=None,
                engine='pyarrow',
            )
            data = data.astype(float)
            if update_index:
                # get number of dofs
                num_dof = int(data.shape[1] / len(self.elements) / 2.0)
                # construct header
                header_data = [
                    (element, station, dof)
                    for element in self.elements
                    for station in (0.00, 1.00)
                    for dof in range(1, num_dof + 1)
                ]
                data.columns = pd.MultiIndex.from_tuples(
                    header_data, names=('element', 'station', 'dof')
                )
                data.index.name = 'time' if self.output_time else None
            self._data = data
        return self._data



"""Defines Model objects."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Literal

import numpy as np

from osmg.core.common import NDM
from osmg.core.gridsystem import GridSystem, GridSystem2D
from osmg.core.osmg_collections import ComponentAssemblyCollection, NodeCollection
from osmg.creators.uid import UIDGenerator

if TYPE_CHECKING:
    from osmg.model_objects.node import Node

from osmg.core.common import numpy_array


@dataclass(repr=False)
class Model:
    """
    Base Model object.

    A general representation of a structural model.

    Attributes:
        name (str): Name of the model.
        grid_system (GridSystem): Grid system of the model.
        uid_generator (UIDGenerator): Object for generating unique IDs.
    """

    name: str
    dimensionality: Literal['2D Truss', '2D Frame', '3D Truss', '3D Frame']
    uid_generator: UIDGenerator = field(default_factory=UIDGenerator)
    nodes: NodeCollection = field(default_factory=NodeCollection)
    components: ComponentAssemblyCollection = field(
        default_factory=ComponentAssemblyCollection
    )

    def bounding_box(self, padding: float) -> tuple[numpy_array, numpy_array]:
        """
        Obtain the axis-aligned bounding box of the building.

        Returns:
          Bounding box.
        """
        num_dimensions = NDM[self.dimensionality]
        p_min = np.full(num_dimensions, np.inf)
        p_max = np.full(num_dimensions, -np.inf)
        for node in list(self.nodes.values()):
            point: numpy_array = np.array(node.coordinates)
            p_min = np.minimum(p_min, point)
            p_max = np.maximum(p_max, point)
        p_min -= np.full(num_dimensions, padding)
        p_max += np.full(num_dimensions, padding)
        return p_min, p_max

    def reference_length(self) -> float:
        """
        Obtain the largest bounding box dimension.

        (used in graphics)

        Returns:
          The largest dimension.
        """
        p_min, p_max = self.bounding_box(padding=0.00)
        return float(np.max(p_max - p_min))

    def get_all_nodes(
        self, ignore_by_tag: set[str] | None = None
    ) -> dict[int, Node]:
        """
        Get all nodes in the model.

        Params:
          ignore_by_tag: Set of tags of components to ignore.

        Returns:
          A dictionary with the nodes. Keys are their UIDs.
        """
        # primary nodes
        primary_nodes = self.nodes
        # internal nodes (of component assemblies)
        internal_nodes: dict[int, Node] = {}
        components = self.components.values()
        for component in components:
            if ignore_by_tag and component.tags & ignore_by_tag:
                continue
            internal_nodes.update(component.internal_nodes)

        all_nodes: dict[int, Node] = {}
        all_nodes.update(primary_nodes)
        all_nodes.update(internal_nodes)
        return all_nodes

    def __repr__(self) -> str:
        """Return a string representation of the object."""
        return f'~~~ Model Object: {self.name} ~~~'


@dataclass(repr=False)
class Model2D(Model):
    """
    2D Model object.

    A 2D model representation.

    Attributes:
        name (str): Name of the model.
        grid_system (GridSystem2D): Grid system for the 2D model.
    """

    dimensionality: Literal['2D Truss', '2D Frame']
    grid_system: GridSystem2D = field(default_factory=GridSystem2D)

    def __post_init__(self) -> None:
        """
        Post-initialization.

        Raises:
          ValueError: If the `dimensionality` assignment is invalid.
        """
        if self.dimensionality not in {'2D Truss', '2D Frame'}:
            msg = f'Dimensionality `{self.dimensionality}` is not compatible with a `Model2D` object.'
            raise ValueError(msg)

    def __repr__(self) -> str:
        """Return a string representation of the object."""
        return f'~~~ 2D Model Object: {self.name} ~~~'


@dataclass(repr=False)
class Model3D(Model):
    """
    3D Model object.

    A 3D model representation.

    Attributes:
        name (str): Name of the model.
        grid_system (GridSystem): Grid system for the 3D model.
    """

    dimensionality: Literal['3D Truss', '3D Frame']
    grid_system: GridSystem = field(default_factory=GridSystem)

    def __post_init__(self) -> None:
        """
        Post-initialization.

        Raises:
          ValueError: If the `dimensionality` assignment is invalid.
        """
        if self.dimensionality not in {'3D Truss', '3D Frame'}:
            msg = f'Dimensionality `{self.dimensionality}` is not compatible with a `Model3D` object.'
            raise ValueError(msg)

    def __repr__(self) -> str:
        """Return a string representation of the object."""
        return f'~~~ 3D Model Object: {self.name} ~~~'
